@@ -47,6 +47,7 @@
 #include "bluetooth-fdo-glue.h"
 #include "bluetooth-utils.h"
 #include "gnome-bluetooth-enum-types.h"
+#include "pin.h"
 
 #define BLUEZ_SERVICE			"org.bluez"
 #define BLUEZ_MANAGER_PATH		"/"
@@ -75,6 +76,13 @@ enum {
 	PROP_DEFAULT_ADAPTER_DISCOVERING
 };
 
+enum {
+	DEVICE_REMOVED,
+	LAST_SIGNAL
+};
+
+static guint signals[LAST_SIGNAL] = { 0 };
+
 static const char *connectable_uuids[] = {
 	"HSP",
 	"AudioSource",
@@ -241,6 +249,54 @@ bluetooth_client_get_connectable(const char **uuids)
 	return FALSE;
 }
 
+static const char *
+phone_oui_to_icon_name (const char *bdaddr)
+{
+	char *vendor;
+	const char *ret = NULL;
+
+	vendor = oui_to_vendor (bdaddr);
+	if (vendor == NULL)
+		return NULL;
+
+	if (strstr (vendor, "Apple") != NULL)
+		ret = "phone-apple-iphone";
+	else if (strstr (vendor, "Samsung") != NULL)
+		ret = "phone-samsung-galaxy-s";
+	else if (strstr (vendor, "Google") != NULL)
+		ret = "phone-google-nexus-one";
+	g_free (vendor);
+
+	return ret;
+}
+
+static const char *
+icon_override (const char    *bdaddr,
+	       BluetoothType  type)
+{
+	/* audio-card, you're ugly */
+	switch (type) {
+	case BLUETOOTH_TYPE_HEADSET:
+		return "audio-headset";
+	case BLUETOOTH_TYPE_HEADPHONES:
+		return "audio-headphones";
+	case BLUETOOTH_TYPE_OTHER_AUDIO:
+		return "audio-speakers";
+	case BLUETOOTH_TYPE_PHONE:
+		return phone_oui_to_icon_name (bdaddr);
+	case BLUETOOTH_TYPE_DISPLAY:
+		return "video-display";
+	case BLUETOOTH_TYPE_SCANNER:
+		return "scanner";
+	case BLUETOOTH_TYPE_REMOTE_CONTROL:
+	case BLUETOOTH_TYPE_WEARABLE:
+	case BLUETOOTH_TYPE_TOY:
+		/* FIXME */
+	default:
+		return NULL;
+	}
+}
+
 static void
 device_g_properties_changed (GDBusProxy      *device,
 			     GVariant        *changed_p,
@@ -272,8 +328,11 @@ device_g_properties_changed (GDBusProxy      *device,
 		} else if (g_str_equal (property, "Icon") == TRUE) {
 			const gchar *icon = g_variant_get_string (v, NULL);
 
-			gtk_tree_store_set (priv->store, &iter,
-					    BLUETOOTH_COLUMN_ICON, icon, -1);
+			/* See "Class" handling below */
+			if (g_strcmp0 (icon, "audio-card") != 0) {
+				gtk_tree_store_set (priv->store, &iter,
+						    BLUETOOTH_COLUMN_ICON, icon, -1);
+			}
 		} else if (g_str_equal (property, "Paired") == TRUE) {
 			gboolean paired = g_variant_get_boolean (v);
 
@@ -296,6 +355,37 @@ device_g_properties_changed (GDBusProxy      *device,
 			gtk_tree_store_set (priv->store, &iter,
 					    BLUETOOTH_COLUMN_UUIDS, uuids, -1);
 			g_strfreev (uuids);
+		} else if (g_str_equal (property, "LegacyPairing") == TRUE) {
+			gboolean legacypairing;
+
+			legacypairing = g_variant_get_boolean (v);
+			gtk_tree_store_set (priv->store, &iter,
+					    BLUETOOTH_COLUMN_LEGACYPAIRING, legacypairing,
+					    -1);
+		} else if (g_str_equal (property, "Class") == TRUE) {
+			BluetoothType type;
+			const char *icon = NULL;
+			char *bdaddr;
+
+			gtk_tree_model_get (GTK_TREE_MODEL (priv->store), &iter,
+					    BLUETOOTH_COLUMN_ADDRESS, &bdaddr,
+					    -1);
+
+			type = v ? bluetooth_class_to_type (g_variant_get_uint32 (v)) : BLUETOOTH_TYPE_ANY;
+			icon = icon_override (bdaddr, type);
+
+			g_free (bdaddr);
+
+			if (icon) {
+				gtk_tree_store_set (priv->store, &iter,
+						    BLUETOOTH_COLUMN_TYPE, type,
+						    BLUETOOTH_COLUMN_ICON, icon,
+						    -1);
+			} else {
+				gtk_tree_store_set (priv->store, &iter,
+						    BLUETOOTH_COLUMN_TYPE, type,
+						    -1);
+			}
 		} else {
 			g_debug ("Unhandled property: %s", property);
 		}
@@ -319,7 +409,7 @@ device_added (ObjectManager   *manager,
 	char **uuids;
 	gboolean paired, trusted, connected;
 	int legacypairing;
-	guint type;
+	BluetoothType type;
 	GtkTreeIter iter, parent;
 
 	device = device1_proxy_new_for_bus_sync (G_BUS_TYPE_SYSTEM,
@@ -355,9 +445,12 @@ device_added (ObjectManager   *manager,
 
 	v = g_variant_lookup_value (dict, "Class", G_VARIANT_TYPE_UINT32);
 	type = v ? bluetooth_class_to_type (g_variant_get_uint32 (v)) : BLUETOOTH_TYPE_ANY;
+	icon = icon_override (address, type);
 
-	v = g_variant_lookup_value (dict, "Icon", G_VARIANT_TYPE_STRING);
-	icon = v ? g_variant_get_string (v, NULL) : "bluetooth";
+	if (icon == NULL) {
+		v = g_variant_lookup_value (dict, "Icon", G_VARIANT_TYPE_STRING);
+		icon = v ? g_variant_get_string (v, NULL) : "bluetooth";
+	}
 
 	v = g_variant_lookup_value (dict, "Paired", G_VARIANT_TYPE_BOOLEAN);
 	paired = v ? g_variant_get_boolean (v) : FALSE;
@@ -382,23 +475,37 @@ device_added (ObjectManager   *manager,
 	gtk_tree_model_get (GTK_TREE_MODEL(priv->store), &parent,
 			    BLUETOOTH_COLUMN_PROXY, &adapter, -1);
 
-	if (get_iter_from_address (priv->store, &iter, address, adapter) == FALSE)
-		gtk_tree_store_insert (priv->store, &iter, &parent, -1);
-
-	gtk_tree_store_set(priv->store, &iter,
-			   BLUETOOTH_COLUMN_ADDRESS, address,
-			   BLUETOOTH_COLUMN_ALIAS, alias,
-			   BLUETOOTH_COLUMN_NAME, name,
-			   BLUETOOTH_COLUMN_TYPE, type,
-			   BLUETOOTH_COLUMN_ICON, icon,
-			   BLUETOOTH_COLUMN_LEGACYPAIRING, legacypairing,
-			   BLUETOOTH_COLUMN_UUIDS, uuids,
-			   BLUETOOTH_COLUMN_PAIRED, paired,
-			   BLUETOOTH_COLUMN_CONNECTED, connected,
-			   BLUETOOTH_COLUMN_TRUSTED, trusted,
-			   BLUETOOTH_COLUMN_PROXY, device,
-			   BLUETOOTH_COLUMN_PROPERTIES, properties,
-			   -1);
+	if (get_iter_from_address (priv->store, &iter, address, adapter) == FALSE) {
+		gtk_tree_store_insert_with_values (priv->store, &iter, &parent, -1,
+						   BLUETOOTH_COLUMN_ADDRESS, address,
+						   BLUETOOTH_COLUMN_ALIAS, alias,
+						   BLUETOOTH_COLUMN_NAME, name,
+						   BLUETOOTH_COLUMN_TYPE, type,
+						   BLUETOOTH_COLUMN_ICON, icon,
+						   BLUETOOTH_COLUMN_LEGACYPAIRING, legacypairing,
+						   BLUETOOTH_COLUMN_UUIDS, uuids,
+						   BLUETOOTH_COLUMN_PAIRED, paired,
+						   BLUETOOTH_COLUMN_CONNECTED, connected,
+						   BLUETOOTH_COLUMN_TRUSTED, trusted,
+						   BLUETOOTH_COLUMN_PROXY, device,
+						   BLUETOOTH_COLUMN_PROPERTIES, properties,
+						   -1);
+	} else {
+		gtk_tree_store_set(priv->store, &iter,
+				   BLUETOOTH_COLUMN_ADDRESS, address,
+				   BLUETOOTH_COLUMN_ALIAS, alias,
+				   BLUETOOTH_COLUMN_NAME, name,
+				   BLUETOOTH_COLUMN_TYPE, type,
+				   BLUETOOTH_COLUMN_ICON, icon,
+				   BLUETOOTH_COLUMN_LEGACYPAIRING, legacypairing,
+				   BLUETOOTH_COLUMN_UUIDS, uuids,
+				   BLUETOOTH_COLUMN_PAIRED, paired,
+				   BLUETOOTH_COLUMN_CONNECTED, connected,
+				   BLUETOOTH_COLUMN_TRUSTED, trusted,
+				   BLUETOOTH_COLUMN_PROXY, device,
+				   BLUETOOTH_COLUMN_PROPERTIES, properties,
+				   -1);
+	}
 	g_strfreev (uuids);
 
 	g_signal_connect (G_OBJECT (device), "g-properties-changed",
@@ -416,8 +523,10 @@ device_removed (const char      *path,
 	BluetoothClientPrivate *priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
 	GtkTreeIter iter;
 
-	if (get_iter_from_path(priv->store, &iter, path) == TRUE)
+	if (get_iter_from_path(priv->store, &iter, path) == TRUE) {
+		g_signal_emit (G_OBJECT (client), signals[DEVICE_REMOVED], 0, path);
 		gtk_tree_store_remove(priv->store, &iter);
+	}
 }
 
 static void
@@ -497,6 +606,8 @@ default_adapter_changed (ObjectManager   *manager,
 		g_object_notify (G_OBJECT (client), "default-adapter");
 		g_object_notify (G_OBJECT (client), "default-adapter-powered");
 		g_object_notify (G_OBJECT (client), "default-adapter-discoverable");
+		g_object_notify (G_OBJECT (client), "default-adapter-discovering");
+		g_object_notify (G_OBJECT (client), "default-adapter-name");
 		return;
 	}
 
@@ -534,14 +645,21 @@ adapter_g_properties_changed (GDBusProxy      *adapter,
 					    BLUETOOTH_COLUMN_NAME, name, -1);
 			gtk_tree_model_get (GTK_TREE_MODEL(priv->store), &iter,
 					    BLUETOOTH_COLUMN_DEFAULT, &is_default, -1);
-			if (is_default != FALSE)
+			if (is_default != FALSE) {
 				g_object_notify (G_OBJECT (client), "default-adapter-powered");
+				g_object_notify (G_OBJECT (client), "default-adapter-name");
+			}
 			notify = TRUE;
 		} else if (g_str_equal (property, "Discovering") == TRUE) {
 			gboolean discovering = g_variant_get_boolean (v);
+			gboolean is_default;
 
 			gtk_tree_store_set (priv->store, &iter,
 					    BLUETOOTH_COLUMN_DISCOVERING, discovering, -1);
+			gtk_tree_model_get (GTK_TREE_MODEL(priv->store), &iter,
+					    BLUETOOTH_COLUMN_DEFAULT, &is_default, -1);
+			if (is_default != FALSE)
+				g_object_notify (G_OBJECT (client), "default-adapter-discovering");
 			notify = TRUE;
 		} else if (g_str_equal (property, "Powered") == TRUE) {
 			gboolean powered = g_variant_get_boolean (v);
@@ -555,6 +673,8 @@ adapter_g_properties_changed (GDBusProxy      *adapter,
 				g_object_notify (G_OBJECT (client), "default-adapter");
 				g_object_notify (G_OBJECT (client), "default-adapter-powered");
 				g_object_notify (G_OBJECT (client), "default-adapter-discoverable");
+				g_object_notify (G_OBJECT (client), "default-adapter-discovering");
+				g_object_notify (G_OBJECT (client), "default-adapter-name");
 			}
 			notify = TRUE;
 		} else if (g_str_equal (property, "Discoverable") == TRUE) {
@@ -685,6 +805,7 @@ adapter_removed (ObjectManager   *manager,
 		g_object_notify (G_OBJECT (client), "default-adapter");
 		g_object_notify (G_OBJECT (client), "default-adapter-powered");
 		g_object_notify (G_OBJECT (client), "default-adapter-discoverable");
+		g_object_notify (G_OBJECT (client), "default-adapter-discovering");
 	}
 }
 
@@ -697,14 +818,14 @@ interface_added (BluetoothClient *client,
 
 	if (g_variant_lookup_value (variant, BLUEZ_ADAPTER_INTERFACE,
 				    G_VARIANT_TYPE_DICTIONARY)) {
-		g_debug ("New Adapter interface added.\n");
+		g_debug ("New Adapter interface added.");
 		adapter_added (priv->manager, path, variant, client);
 		return;
 	}
 
 	if (g_variant_lookup_value (variant, BLUEZ_DEVICE_INTERFACE,
 				    G_VARIANT_TYPE_DICTIONARY)) {
-		g_debug ("New Device interface added.\n");
+		g_debug ("New Device interface added.");
 		device_added (priv->manager, client, path, variant);
 		return;
 	}
@@ -947,20 +1068,21 @@ _bluetooth_client_set_discoverable (BluetoothClient *client,
 {
 	BluetoothClientPrivate *priv = BLUETOOTH_CLIENT_GET_PRIVATE (client);
 	GError *error = NULL;
-	GDBusProxy *adapter;
+	GtkTreePath *path;
 	Properties *properties;
 	gboolean ret;
 	GtkTreeIter iter;
 
 	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), FALSE);
 
-	adapter = _bluetooth_client_get_default_adapter (client);
-	if (adapter == NULL)
+	if (priv->default_adapter == NULL)
 		return FALSE;
 
-	get_iter_from_proxy (priv->store, &iter, adapter);
+	path = gtk_tree_row_reference_get_path (priv->default_adapter);
+	gtk_tree_model_get_iter (GTK_TREE_MODEL (priv->store), &iter, path);
 	gtk_tree_model_get (GTK_TREE_MODEL (priv->store), &iter,
-			    BLUETOOTH_COLUMN_PROPERTIES, &properties, -1);
+                            BLUETOOTH_COLUMN_PROPERTIES, &properties, -1);
+        gtk_tree_path_free (path);
 
 	if (properties == NULL)
 		return FALSE;
@@ -973,10 +1095,19 @@ _bluetooth_client_set_discoverable (BluetoothClient *client,
 	if (ret == FALSE) {
 		g_warning ("Failed to set Discoverable to %d: %s", discoverable, error->message);
 		g_error_free (error);
+	} else {
+		ret = properties_call_set_sync (properties,
+						BLUEZ_ADAPTER_INTERFACE,
+						"DiscoverableTimeout",
+						g_variant_new_variant (g_variant_new_uint32 (timeout)),
+						NULL, &error);
+		if (ret == FALSE) {
+			g_warning ("Failed to set DiscoverableTimeout to %d: %s", timeout, error->message);
+			g_error_free (error);
+		}
 	}
 
 	g_object_unref (properties);
-	g_object_unref(adapter);
 
 	return ret;
 }
@@ -1105,6 +1236,23 @@ static void bluetooth_client_class_init(BluetoothClientClass *klass)
 	object_class->get_property = bluetooth_client_get_property;
 	object_class->set_property = bluetooth_client_set_property;
 
+	/**
+	 * BluetoothClient::device-removed:
+	 * @client: a #BluetoothClient object which received the signal
+	 * @device: the D-Bus object path for the now-removed device
+	 *
+	 * The #BluetoothClient::device-removed signal is launched when a
+	 * device gets removed from the model.
+	 **/
+	signals[DEVICE_REMOVED] =
+		g_signal_new ("device-removed",
+			      G_TYPE_FROM_CLASS (klass),
+			      G_SIGNAL_RUN_LAST,
+			      0,
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__STRING,
+			      G_TYPE_NONE, 1, G_TYPE_STRING);
+
 	/**
 	 * BluetoothClient:default-adapter:
 	 *
@@ -1310,55 +1458,91 @@ typedef struct {
 } CreateDeviceData;
 
 static void
-device_pair_callback (GDBusProxy       *proxy,
-		      GAsyncResult     *res,
-		      CreateDeviceData *devdata)
+device_pair_callback (GDBusProxy         *proxy,
+		      GAsyncResult       *res,
+		      GSimpleAsyncResult *simple)
 {
 	GError *error = NULL;
-	const char *path;
 
-	if (device1_call_pair_finish (DEVICE1(proxy), res, &error) == FALSE)
-		g_warning ("Pair() failed: %s", error->message);
+	if (device1_call_pair_finish (DEVICE1(proxy), res, &error) == FALSE) {
+		g_debug ("Pair() failed for %s: %s",
+			 g_dbus_proxy_get_object_path (proxy),
+			 error->message);
+		g_simple_async_result_take_error (simple, error);
+	} else {
+		g_simple_async_result_set_op_res_gboolean (simple, TRUE);
+	}
 
-	path = g_dbus_proxy_get_object_path(proxy);
-	if (devdata->func)
-		devdata->func (devdata->client, error, path);
+	g_simple_async_result_complete_in_idle (simple);
 
-	if (error != NULL)
-		g_error_free (error);
-
-	g_object_unref (devdata->client);
-	g_free (devdata);
+	g_object_unref (simple);
 }
 
 gboolean
+bluetooth_client_setup_device_finish (BluetoothClient  *client,
+				      GAsyncResult     *res,
+				      char            **path,
+				      GError          **error)
+{
+	GSimpleAsyncResult *simple;
+
+	simple = (GSimpleAsyncResult *) res;
+
+	g_warn_if_fail (g_simple_async_result_get_source_tag (simple) == bluetooth_client_setup_device);
+
+	if (path != NULL)
+		*path = g_object_get_data (G_OBJECT (res), "device-object-path");
+
+	if (g_simple_async_result_get_op_res_gboolean (simple))
+		return TRUE;
+	g_simple_async_result_propagate_error (simple, error);
+	return FALSE;
+}
+
+void
 bluetooth_client_setup_device (BluetoothClient          *client,
 			       const char               *path,
-			       const char               *agent,
-			       BluetoothClientSetupFunc  func,
-			       gboolean                  pair)
+			       gboolean                  pair,
+			       GCancellable             *cancellable,
+			       GAsyncReadyCallback       callback,
+			       gpointer                  user_data)
 {
 	BluetoothClientPrivate *priv = BLUETOOTH_CLIENT_GET_PRIVATE(client);
-	CreateDeviceData *devdata;
-	GDBusProxy *adapter, *device;
+	GSimpleAsyncResult *simple;
+	GDBusProxy *device;
 	GtkTreeIter iter, adapter_iter;
 	gboolean paired;
 	GError *err = NULL;
 
 	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), FALSE);
 
-	if (get_iter_from_path (priv->store, &iter, path) == FALSE)
-		return FALSE;
+	simple = g_simple_async_result_new (G_OBJECT (client),
+					    callback,
+					    user_data,
+					    bluetooth_client_setup_device);
+	g_simple_async_result_set_check_cancellable (simple, cancellable);
+	g_object_set_data (G_OBJECT (simple), "device-object-path", g_strdup (path));
+
+	if (get_iter_from_path (priv->store, &iter, path) == FALSE) {
+		g_simple_async_result_set_error (simple, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
+						 "Device with object path %s does not exist",
+						 path);
+		g_simple_async_result_complete_in_idle (simple);
+		g_object_unref (simple);
+		return;
+	}
 
 	gtk_tree_model_get (GTK_TREE_MODEL(priv->store), &iter,
 			    BLUETOOTH_COLUMN_PROXY, &device,
 			    BLUETOOTH_COLUMN_PAIRED, &paired, -1);
 
 	if (paired != FALSE &&
-			gtk_tree_model_iter_parent (GTK_TREE_MODEL(priv->store),
-						    &adapter_iter, &iter)) {
+	    gtk_tree_model_iter_parent (GTK_TREE_MODEL(priv->store), &adapter_iter, &iter)) {
+		GDBusProxy *adapter;
+
 		gtk_tree_model_get (GTK_TREE_MODEL(priv->store), &adapter_iter,
-				    BLUETOOTH_COLUMN_PROXY, &adapter, -1);
+				    BLUETOOTH_COLUMN_PROXY, &adapter,
+				    -1);
 		adapter1_call_remove_device_sync (ADAPTER1 (adapter),
 						  path,
 						  NULL, &err);
@@ -1370,21 +1554,17 @@ bluetooth_client_setup_device (BluetoothClient          *client,
 	}
 
 	if (pair == TRUE) {
-		devdata = g_new0 (CreateDeviceData, 1);
-		devdata->func = func;
-		devdata->client = g_object_ref (client);
-
 		device1_call_pair (DEVICE1(device),
-				   NULL,
+				   cancellable,
 				   (GAsyncReadyCallback) device_pair_callback,
-				   devdata);
+				   simple);
 	} else {
-		if (func)
-			func (client, NULL, path);
+		g_simple_async_result_set_op_res_gboolean (simple, TRUE);
+		g_simple_async_result_complete_in_idle (simple);
+		g_object_unref (simple);
 	}
 
 	g_object_unref (device);
-	return TRUE;
 }
 
 gboolean
@@ -1401,14 +1581,18 @@ bluetooth_client_set_trusted (BluetoothClient *client,
 	g_return_val_if_fail (BLUETOOTH_IS_CLIENT (client), FALSE);
 	g_return_val_if_fail (device != NULL, FALSE);
 
-	if (get_iter_from_path (priv->store, &iter, device) == FALSE)
+	if (get_iter_from_path (priv->store, &iter, device) == FALSE) {
+		g_debug ("Couldn't find device '%s' in tree to mark it as trusted", device);
 		return FALSE;
+	}
 
 	gtk_tree_model_get (GTK_TREE_MODEL (priv->store), &iter,
 			    BLUETOOTH_COLUMN_PROPERTIES, &properties, -1);
 
-	if (properties == NULL)
+	if (properties == NULL) {
+		g_debug ("Couldn't find properties for device '%s' in tree to mark it as trusted", device);
 		return FALSE;
+	}
 
 	ret = properties_call_set_sync (properties, BLUEZ_DEVICE_INTERFACE, "Trusted",
 					g_variant_new_variant (g_variant_new_boolean (trusted)),
@@ -1441,8 +1625,8 @@ bluetooth_client_get_device (BluetoothClient *client,
 }
 
 static void
-connect_callback (GDBusProxy   *proxy,
-		  GAsyncResult *res,
+connect_callback (GDBusProxy         *proxy,
+		  GAsyncResult       *res,
 		  GSimpleAsyncResult *simple)
 {
 	gboolean retval;
@@ -1583,7 +1767,7 @@ bluetooth_client_dump_device (GtkTreeModel *model,
 	char *address, *alias, *name, *icon, **uuids;
 	gboolean is_default, paired, trusted, connected, discoverable, discovering, powered, is_adapter;
 	GtkTreeIter parent;
-	guint type;
+	BluetoothType type;
 
 	gtk_tree_model_get (model, iter,
 			    BLUETOOTH_COLUMN_ADDRESS, &address,
